# -*- coding: UTF-8 -*-
from .__common__ import hashtype, API


__all__ = ["MalwareBazaar"]


class MalwareBazaar(API):
    doc = "https://bazaar.abuse.ch/api"
    url = "https://mb-api.abuse.ch/api/v1"
    _api_key_header = "Auth-Key"
    _post_condition = lambda s: s.json.get('query_status') != "file_not_found"

    @hashtype("sha256")
    def get_file_by_hash(self, hash):
        # daily limit: 2000/IP/day
        self._post("", data={'query': "get_file", 'sha256_hash': hash})._unzip(b"infected")._save()




# -*- coding: UTF-8 -*-
import logging
from os import cpu_count

from .clients import *
from .clients import __all__ as _clients
from .clients.__common__ import _valid_hash


__all__ = ["download_sample", "download_samples", "get_samples_feed"] + _clients

_CLIENTS_MAP = {n.lower(): globals()[n] for n in _clients}
_MAX_WORKERS = 3 * cpu_count()

logger = logging.getLogger("malsearch")


def _check_conf(method):
    def _wrapper(f):
        from functools import wraps
        @wraps(f)
        def _subwrapper(*args, config=None, **kwargs):
            if config is None:
                logger.error("no configuration file provided")
                logger.info(f"you can create one at {config} manually (INI format with section 'API keys')")
            else:
                if isinstance(config, str):
                    config = _valid_conf(config)
                cmap = {n: cls for n, cls in _CLIENTS_MAP.items() if (s := kwargs.get('select')) is None or n in s}
                clients, skipped = [], kwargs.get('skip') or []
                for n in config['API keys']:
                    if n not in cmap.keys() or not hasattr(cmap[n], method):
                        continue
                    if n in skipped:
                        logger.debug(f"{n} skipped")
                        continue
                    if config.has_section("Disabled"):
                        t = config['Disabled'].get(n)
                        if t is not None:
                            import datetime as dt
                            try:
                                if dt.datetime.strptime(t, "%d/%m/%Y %H:%M:%S") < dt.datetime.now():
                                    from contextlib import nullcontext
                                    with kwargs.get('lock') or nullcontext():
                                        config['Disabled'].pop(n)
                                        with open(config.path, 'w') as cfg:
                                            config.write(cfg)
                                else:
                                    logger.warning(f"{n} is disabled until {t}")
                                    continue
                            except ValueError:
                                logger.warning(f"{n} is disabled")
                                continue
                    cls = _CLIENTS_MAP[n]
                    if cls.__base__.__name__ == "API":
                        kwargs['api_key'] = config['API keys'].get(n)
                    clients.append(cls(config=config, **kwargs))
                if len(clients) == 0:
                    logger.warning("no download client available/enabled")
                logger.debug(f"clients: {', '.join(c.name for c in clients)}")
                return f(*args, clients=clients, config=config, **kwargs)
        return _subwrapper
    return _wrapper


def _valid_conf(path):
    from configparser import ConfigParser
    from os.path import exists, expanduser
    path = expanduser(path)
    if not exists(path):
        raise ValueError("configuration file does not exist")
    conf = ConfigParser()
    try:
        conf.read(path)
        conf.path = path
    except:
        raise ValueError("invalid configuration file")
    return conf


def _valid_hash_file(path):
    from os.path import exists, expanduser
    path = expanduser(path)
    if not exists(path):
        raise ValueError("hashes file does not exist")
    hashes = []
    with open(path) as f:
        for l in f:
            hashes.append(_valid_hash(l.strip()))
    return hashes


@_check_conf("get_file_by_hash")
def download_sample(hash, config=None, **kwargs):
    """ Function to download a single sample """
    from os.path import exists, join
    p = join(kwargs.get('output_dir', "."), hash)
    if exists(p) and not kwargs.get('overwrite'):
        logger.info(f"'{p}' already exists")
        return
    for client in kwargs['clients']:
        logger.debug(f"trying {client.name}...")
        try:
            client.get_file_by_hash(hash)
            if len(getattr(client, "content", "")) > 0:
                logger.info(f"found {hash} on {client.__class__.__name__} !")
                return client
        except AttributeError:
            continue  # not a client for downloading samples (e.g. Maldatabase)
        except (HashNotFoundError, ServiceUnavailable, ValueError) as e:
            logger.debug(e)
        except Exception as e:
            logger.exception(e)
    logger.warning(f"could not find {hash}")


def download_samples(*hashes, max_workers=_MAX_WORKERS, **kwargs):
    """ Threaded function to download samples from hashes in parallel. """
    from concurrent.futures import as_completed, ThreadPoolExecutor as Pool
    from threading import Lock
    try:
        from tqdm import tqdm
        TQDM = True
    except ImportError:
        TQDM = False
    kwargs['lock'] = Lock()
    if len(hashes) == 0:
        return
    elif len(hashes) == 1:
        download_sample(hashes[0].lower(), **kwargs)
    else:
        with Pool(max_workers=max_workers) as executor:
            tasks = []
            for h in hashes:
                tasks.append(executor.submit(download_sample, h.lower(), **kwargs))
            for task in (tqdm(as_completed(tasks), total=len(hashes)) if TQDM else as_completed(tasks)):
                task.result()


@_check_conf("get_malware_feed")
def get_samples_feed(config=None, **kwargs):
    count = 0
    for client in kwargs['clients']:
        logger.debug(f"trying {client.name}...")
        try:
            for h in client.get_malware_feed():
                yield h
                count += 1
        except Exception as e:
            logger.exception(e)
    logger.info(f"got {count} hashes")




#!/usr/bin/python
# -*- coding: UTF-8 -*-
import os
from malsearch import *
from unittest import TestCase


CONF = """[API keys]
VirusTotal = 9f86d081884c7d659a2feaa0c55ad015a3bf4f1b2b0b822cd15d6c15b0f00a08
[Disabled]
VirusTotal = Bad API key
"""
FILE = "test.conf"
HASH = "098f6bcd4621d373cade4e832627b4f6"


class TestInit(TestCase):
    @classmethod
    def setUpClass(cls):
        with open(FILE, 'w') as f:
            f.write(CONF)
    
    @classmethod
    def tearDownClass(cls):
        os.remove(FILE)
    
    def test_config(self):
        from malsearch.__init__ import _valid_conf
        self.assertRaises(ValueError, _valid_conf, "does_not_exist")
        self.assertIsNotNone(_valid_conf(FILE))
    
    def test_download(self):
        self.assertIsNone(download_sample(HASH))
        self.assertIsNone(download_sample(HASH, FILE))
        self.assertIsNone(download_samples())




# -*- coding: UTF-8 -*-
import re
import threading

from .__common__ import hashtype, API


__all__ = ["MalwareFeed"]


class MalwareFeed(API):
    doc = "https://github.com/MalwareSamples/Malware-Feed/blob/master/README.md"
    url = "https://api.github.com/repos/MalwareSamples/Malware-Feed"
    _auth_method = "Bearer"
    _pre_condition = lambda s: s.hash in s._paths.keys()
    _headers = {'Accept': "application/vnd.github+json"}
    
    def get_malware_feed(self):
        with threading.Lock() as lock:
            MalwareFeed._paths = {}
            self._get("branches/master")
            self._get(f"git/commits/{self.json['commit']['sha']}")
            self._get(f"git/trees/{self.json['tree']['sha']}", params={'recursive': 1})
            for item in self.json['tree']:
                h = item['path'].split("/")[-1]
                if re.match(r"^[0-9a-f]{64}$", h, re.I):
                    MalwareFeed._paths[h] = item['path']
    
    @hashtype("sha256")
    def get_file_by_hash(self, hash):
        try:
            MalwareFeed._paths
        except AttributeError:
            self.get_malware_feed()
        self._get(f"contents/{MalwareFeed._paths.get(hash)}")._decode("base64")._save()




# -*- coding: UTF-8 -*-
import importlib
import os


__all__ = []

DOCSTR = """
>>> from malsearch import {cls}
>>> {cls}(api_key="YOUR-KEY", error=True).get_file_by_hash("HASH")
"""


for f in sorted(os.listdir(os.path.dirname(os.path.abspath(__file__)))):
    if not f.endswith(".py") or f.startswith("_"):
        continue
    for cls in (module := importlib.import_module(f".{f[:-3]}", package=__name__)).__all__:
        __all__.append(cls)
        globals()[cls] = c = getattr(module, cls)
        c.__doc__ = DOCSTR.format(cls=cls)




# -*- coding: UTF-8 -*-
from .__info__ import __author__, __copyright__, __email__, __license__, __source__, __version__


def _parser(name, description, examples):
    from argparse import ArgumentParser, RawTextHelpFormatter
    descr = f"{name} {__version__}\n\nAuthor   : {__author__} ({__email__})\nCopyright: {__copyright__}\nLicense  :" \
            f" {__license__}\nSource   : {__source__}\n\n{description}.\n\n"
    examples = [f"malsearch {e}" if not e.startswith("malsearch ") else e for e in examples]
    return ArgumentParser(description=descr, formatter_class=RawTextHelpFormatter, add_help=False,
                          epilog="usage examples:\n  " + "\n  ".join(examples) if len(examples) > 0 else None)


def _setup(parser):
    args = parser.parse_args()
    if hasattr(args, "verbose"):
        import logging
        logging.basicConfig(level=[logging.INFO, logging.DEBUG][args.verbose])
    if not args.disable_cache:
        from requests_cache import install_cache
        install_cache("malsearch",
            use_cache_dir=True,                 # Save files in the default user cache dir
            cache_control=True,                 # Use Cache-Control response headers for expiration, if available
            expire_after=86400,                 # Otherwise expire responses after one day
            allowable_codes=[404],              # Cache 400 responses as a solemn reminder of your failures
            allowable_methods=['GET', 'POST'],  # Cache whatever HTTP methods you want
            ignored_parameters=['api_key'],     # Don't match this request param, and redact if from the cache
            match_headers=['Accept-Language'],  # Cache a different response per language
            stale_if_error=True,                # In case of request errors, use stale cache data if possible
        )
    return args


def main():
    from os import makedirs
    from .__init__ import _valid_conf, _valid_hash, _valid_hash_file, download_samples, get_samples_feed
    parser = _parser("MalSearch", "This tool is aimed to search for malware samples across some public databases",
                     ["2037f9b7dd268eef7d2e950b27c6cf80e3ba692d262c785ab67b04dc71c99bf9",
                      "094fd325049b8a9cf6d3e5ef2a6d4cc6a567d7d49c35f8bb8dd9e3c6acf3d78d --select malwarebazaar",
                      "-f hashes.txt -o samples --disable-cache"])
    parser.add_argument("sample_hash", type=_valid_hash, nargs="*", help="input hash")
    parser.add_argument("-f", "--from-file", type=_valid_hash_file,
                        help="get hashes from the target file (newline-separated list)")
    parser.add_argument("-m", "--from-malware-feed", action="store_true",
                        help="get hashes from malware feeds (default: False)")
    opt = parser.add_argument_group("optional arguments")
    opt.add_argument("-c", "--config", default="~/.malsearch.conf", type=_valid_conf,
                     help="INI configuration file (default: ~/.malsearch.conf)")
    opt.add_argument("-o", "--output-dir", default=".", help="output directory for downloaded samples (default: .)")
    opt_clients = opt.add_mutually_exclusive_group()
    opt_clients.add_argument("--select", nargs="*",
                             help="select the specified clients while downloading samples (default: all)")
    opt_clients.add_argument("--skip", nargs="*",
                             help="skip the specified clients while downloading samples (default: none)")
    opt.add_argument("-u", "--unpacked", action="store_true",
                     help="if available and target sample is packed, download unpacked version too")
    opt.add_argument("--disable-cache", action="store_true", help="disable requests cache")
    extra = parser.add_argument_group("extra arguments")
    extra.add_argument("-h", "--help", action="help", help="show this help message and exit")
    extra.add_argument("-v", "--verbose", action="store_true", help="display debug information (default: False)")
    args = _setup(parser)
    if args.from_file is not None:
        with open(args.from_file) as f:
            for h in f.readlines():
                args.sample_hash.append(_valid_hash(h.strip()))
    if args.from_malware_feed:
        for h in get_samples_feed():
            args.sample_hash.append(_valid_hash(h.strip()))
    makedirs(args.output_dir, exist_ok=True)
    if len(args.sample_hash) > 0:
        download_samples(*args.sample_hash, **vars(args))
    else:
        import logging
        logging.getLogger("malsearch").warning("nothing to download")




# -*- coding: UTF-8 -*-
import builtins
import logging


__all__ = ["API", "Web"]

_HASHTYPES_MAP = {'md5': 32, 'sha1': 40, 'sha224': 56, 'sha256': 64, 'sha384': 96, 'sha512': 128}
_MIN_BACKOFF = 300

logger = logging.getLogger("malsearch")


def _hash_type(hash, *types):
    try:
        return {v: k for k, v in _HASHTYPES_MAP.items()}[len(_valid_hash(hash))]
    except ValueError:
        pass


def _valid_hash(hash, *types):
    import re
    pattern = "|".join(f"[0-9a-f]{{{_HASHTYPES_MAP[t]}}}" for t in (types or _HASHTYPES_MAP.keys()))
    if re.match(r"{}".format(f"^{pattern}$"), hash, re.I) is None:
        raise ValueError("hash type not supported")
    return hash.lower()


def hashtype(*types):
    def _wrapper(f):
        from functools import wraps
        @wraps(f)
        def _subwrapper(*args, **kwargs):
            _valid_hash(args[0] if isinstance(args[0], str) else args[1], *types)
            if not isinstance(args[0], str):
                args[0].hash = args[1]
            return f(*args, **kwargs)
        return _subwrapper
    return _wrapper


class HashNotFoundError(Exception):
    __module__ = 'builtins'
    
    def __init__(self, hash, name):            
        super().__init__(f"{hash} is not known to {name}")
builtins.HashNotFoundError = HashNotFoundError


class ServiceUnavailable(Exception):
    __module__ = 'builtins'
    
    def __init__(self, name, reason=None):            
        super().__init__(f"{name} seems to be currently unavailable{[f' ({reason})', ''][reason is None]}")
builtins.ServiceUnavailable = ServiceUnavailable


class _Base:
    def __init__(self, **kwargs):
        self.name = self.__class__.__name__.lower()
        self._error = False
        self._output_dir = "./"
        self._postcheck = True
        for k, v in kwargs.items():
            setattr(self, f"_{k}", v)
    
    def _decode(self, codec):
        try:
            import codext as codecs
        except ImportError:
            import codecs
        try:
            self.content = self.content.encode()
        except:
            pass
        try:
            self.content = codecs.decode(self.content, codec)
        except ValueError as e:
            [logger.debug, logger.error][self._error](f"cannot decode ({e})")
        except Exception as e:
            logger.exception(e)
        return self
    
    def _save(self, filename=None):
        from os.path import join
        try:
            c = self.content
            with open(join(self._output_dir, filename or self.hash), 'wb') as f:
                f.write(c)
            if filename is None and self._postcheck:
                import hashlib
                h = getattr(hashlib, _hash_type(self.hash))()
                h.update(self.content)
                if (hd := h.hexdigest()) != self.hash:
                    logger.warning(f"content and hash mismatch ({hd} != {self.hash})")
        except AttributeError:
            if self.enabled:
                [logger.debug, logger.error][self._error]("cannot save (no content downloaded)")
        except Exception as e:
            logger.exception(e)
        return self
    
    def _unzip(self, password=None):
        from io import BytesIO
        try:
            with BytesIO(self.content) as buffer:
                from pyzipper import AESZipFile
                from zipfile import ZipFile
                for ZipClass in [ZipFile, AESZipFile]:
                    try:
                        with ZipClass(buffer) as zip_file:
                            if password:
                                zip_file.setpassword(password)
                            with zip_file.open(zip_file.namelist()[0]) as file:
                                self.content = file.read()
                                logger.debug("unzipped with {ZipClass}")
                                break
                    except NotImplementedError:
                        continue
        except AttributeError:
            if self.enabled:
                [logger.debug, logger.error][self._error]("cannot unzip (no content downloaded)")
        except Exception as e:
            logger.exception(e)
        return self
    
    @property
    def enabled(self):
        return not hasattr(self, "_config") or \
               not self._config.has_section("Disabled") or \
               self.name not in self._config['Disabled']


class API(_Base):
    def __request(self, path, **kwargs):
        from requests import exceptions, get, post
        if not getattr(self, "_pre_condition", lambda: True)():
            raise HashNotFoundError(self.hash, self.__class__.__name__)
        for i in ["headers", "params"]:
            for k, v in getattr(self, f"_{i}", {}).items():
                kwargs[i] = kwargs[i] or {}
                kwargs[i].setdefault(k, v)
            if n := getattr(self, f"_api_key_{i.rstrip('s')}", None):
                d = kwargs.pop(i, {}) or {}
                d[n] = self._api_key
                kwargs[i] = d
        if n := getattr(self, "_auth_method", None):
            kwargs['headers'] = kwargs['headers'] or {}
            kwargs['headers']['Authorization'] = f"{n} {self._api_key}"
        r = (get if 'params' in kwargs else post)(f"{self.url}/{path}", **kwargs)
        try:
            r.raise_for_status()
            self.content = r.content
        except exceptions.RequestException as e:
            if r.status_code == 403:
                import datetime as dt
                from contextlib import nullcontext
                if hasattr(self, "_config"):
                    with kwargs.get('lock') or nullcontext():
                        if not self._config.has_section("Disabled"):
                            self._config.add_section("Disabled")
                        boff = dt.timedelta(seconds=getattr(self, "backoff", _MIN_BACKOFF))
                        self._config['Disabled'][self.name] = dt.datetime.strftime(dt.datetime.now() + boff,
                                                                                   "%d/%m/%Y %H:%M:%S")
                        with open(self._config.path, 'w') as f:
                            self._config.write(f)
                    logger.warning(f"{self.name} raised 403 (Forbidden) and was disabled ; check out your API key")
            elif r.status_code == 404:
                raise HashNotFoundError(self.hash, self.__class__.__name__)
            elif 500 <= r.status_code < 600:
                raise ServiceUnavailable(self.__class__.__name__, f"{r.status_code} - {r.reason}")
            else:
                logger.exception(e)
        try:
            self.json = r.json()
            if 'content' in self.json.keys():
                self.content = self.json['content']
        except:
            pass
        if not getattr(self, "_post_condition", lambda: True)():
            raise HashNotFoundError(self.hash, self.__class__.__name__)

    def _get(self, path, params=None, headers=None):
        """ Perform a GET request.
        
        :param path:    API endpoint (relative path).
        :param params:  Dictionary of URL parameters.
        :param headers: Optional dictionary of request headers.
        :return:        JSON data or error message.
        """
        self.__request(path, params=params, headers=headers)
        return self
    
    def _post(self, path, data=None, json=None, headers=None):
        """ Perform a POST request.

        :param path:    API endpoint (relative path).
        :param data:    Dictionary or list of tuples for form-encoded data.
        :param json:    Dictionary to send JSON data.
        :param headers: Optional dictionary of request headers.
        :return:        JSON data or error message.
        """
        self.__request(path, data=data, json=json, headers=headers)
        return self


class Web(_Base):
    def _get(self, path, params=None, headers=None):
        """ Perform a GET request.
        
        :param path:    API endpoint (relative path).
        :param params:  Dictionary of URL parameters.
        :param headers: Optional dictionary of request headers.
        :return:        BeautifulSoup-parsed HTML object or None.
        """
        from bs4 import BeautifulSoup
        from requests import exceptions, get
        try:
            r = get(path, params=params, headers=headers)
            r.raise_for_status()
            return BeautifulSoup(r.content, "html.parser")
        except exceptions.RequestException as e:
            return




# -*- coding: UTF-8 -*-
from .__common__ import hashtype, API


__all__ = ["VirusShare"]


class VirusShare(API):
    doc = "https://virusshare.com/apiv2_reference"
    url = "https://virusshare.com/apiv2"
    _api_key_param = "apikey"
    
    @hashtype("md5", "sha1", "sha224", "sha256", "sha384", "sha512")
    def get_file_by_hash(self, hash):
         self._get("download", params={'hash': hash})._unzip(b"infected")._save()
 



# -*- coding: UTF-8 -*-
from .__common__ import hashtype, API


__all__ = ["VirusTotal"]


class VirusTotal(API):
    doc = "https://docs.virustotal.com/reference/overview"
    url = "https://www.virustotal.com/api/v3"
    _api_key_header = "x-apikey"
    
    @hashtype("md5", "sha1", "sha256")
    def get_file_by_hash(self, hash):
        if self._unpacked:
            data = self._get(f"/files/{hash}")
            for n in data['data']['attributes']['names']:
                if n.endswith("_unpacked"):
                    hash2, _ = n.split("_", 1)
                    self._get(f"/files/{hash2}/download")._save(hash2)
        self._get(f"files/{hash}/download")._unzip(b"infected")._save()




# -*- coding: UTF-8 -*-
from .__common__ import hashtype, API


__all__ = ["MalShare"]


class MalShare(API):
    doc = "https://malshare.com/doc.php"
    url = "https://malshare.com"
    _api_key_param = "api_key"
    
    @hashtype("md5", "sha1", "sha256")
    def get_file_by_hash(self, hash):
        self._get("api.php", params={'action': "getfile", 'hash': hash})._save()




# -*- coding: UTF-8 -*-
from .__common__ import hashtype, API


__all__ = ["Malpedia"]


class Malpedia(API):
    doc = "https://malpedia.caad.fkie.fraunhofer.de/usage/api"
    url = "https://malpedia.caad.fkie.fraunhofer.de/api"
    _auth_method = "APIToken"
    
    @hashtype("md5", "sha256")
    def get_file_by_hash(self, hash):
        self._get(f"get/sample/{hash}/zip")._save()
        #TODO: parse ZIP for unpacked version




# -*- coding: UTF-8 -*-
from .__common__ import hashtype, API


__all__ = ["Triage"]


class Triage(API):
    doc = "https://tria.ge/docs"
    url = "https://tria.ge/api/v0"
    _auth_method = "Bearer"
    
    @hashtype("md5", "sha1", "sha256", "sha512")
    def get_file_by_hash(self, hash):
        hashtype = {32: "md5", 40: "sha1", 64: "sha256", 128: "sha512"}[len(hash)]
        self._get("search", params={'query': f"{hashtype}:{hash}"})
        try:
            sample_id = self.json['data'][0]['id']
            self._get(f"samples/{sample_id}/sample")._save()
        except IndexError:
            delattr(self, "content")




# -*- coding: UTF-8 -*-
from .__common__ import Web


__all__ = ["VxUnderground"]


class VxUnderground(Web):
    url = "https://vx-underground.org/Samples"
    
    def get_malware_feed(self):
        #TODO
        pass




# -*- coding: UTF-8 -*-
from .__common__ import API


__all__ = ["Maldatabase"]


class Maldatabase(API):
    doc = "https://maldatabase.com/api-doc.html"
    url = "https://api.maldatabase.com/download"
    _api_key_header = "Authorization"
    
    def get_malware_feed(self, hashtype="sha256"):
        # available output hash types: md5, sha1, sha256
        self._get("", headers={'Accept-Encoding': "gzip, deflate"})
        for data in self.json:
            yield data[hashtype]




#!/usr/bin/python
# -*- coding: UTF-8 -*-
import os
from malsearch.clients.__common__ import _valid_hash, hashtype, _Base
from unittest import TestCase

FILE = "tmp-file"
TEST_HASHES = {
    'md5':    "098f6bcd4621d373cade4e832627b4f6",
    'sha1':   "a94a8fe5ccb19ba61c4c0873d391e987982fbbd3",
    'sha224': "90a3ed9e32b2aaf4c61c410eb925426119e1a9dc53d4286ade99a809",
    'sha256': "9f86d081884c7d659a2feaa0c55ad015a3bf4f1b2b0b822cd15d6c15b0f00a08",
    'sha284': "768412320f7b0aa5812fce428dc4706b3cae50e02a64caa16a782249bfe8efc4b7ef1ccb126255d196047dfedf17a0a9",
    'sha512': "ee26b0dd4af7e749aa1a8ee3c10ae9923f618980772e473f8819a5d4940e0db27ac185f8a0e1d5f84f88bc887fd67b143732c304cc5fa9ad8e6f57f50028a8ff",
}


class TestClient(_Base):
    pass


class TestInit(TestCase):
    def test_base_client_class(self):
        c = TestClient(test="ok")
        self.assertEqual(c.name, "testclient")
        self.assertEqual(getattr(c, "_test", None), "ok")
        c.content = b"dGVzdA"                # will fail with logging message
        self.assertIsNotNone(c._decode("base64"))
        c.content = b"dGVzdA=="              # will succeed
        self.assertIsNotNone(c._decode("base64"))
        self.assertEqual(c.content, b"test")
        self.assertIsNotNone(c._unzip())     # will fail with logging message
        self.assertIsNotNone(c._save(FILE))  # will succeed
        delattr(c, "content")
        self.assertIsNotNone(c._save(FILE))  # will fail with logging message
        self.assertIsNotNone(c._unzip())     # will fail with logging message
    
    def test_hash_functions(self):
        @hashtype("sha1", "sha256")
        def test_func(hash):
            pass
        self.assertRaises(ValueError, _valid_hash, "bad")
        for h in TEST_HASHES.values():
            self.assertEqual(_valid_hash(h), h.lower())
        self.assertRaises(ValueError, test_func, "bad")
        self.assertIsNone(test_func(TEST_HASHES['sha1']))
        self.assertRaises(ValueError, test_func, TEST_HASHES['md5'])
        self.assertIsNone(test_func(TEST_HASHES['sha256']))

